extern crate image;
extern crate clap;
#[macro_use]
extern crate colour;

mod reskit;
use std::error::Error;
use std::fs::File;
use std::io::Write;
use clap::{App, Arg, SubCommand};
use reskit::soundtrack::engines::echo::EchoFormat;
use reskit::soundtrack::formats::dmf::DmfModule;
use reskit::utility;
use reskit::tileset;

use crate::reskit::utility::print_good;

fn run() -> Result<(), Box<dyn Error>> {
    let matches = App::new( "reskit" )
        .version( env!( "CARGO_PKG_VERSION" ) )
        .author( "(c) 2021-2023 Ashley N. <ne0ndrag0n@ne0ndrag0n.com>" )
        .about( "The Retro Entertainment Software Toolkit - Tools for authoring software for legacy video game systems" )
        .subcommand(
            SubCommand::with_name( "tileset" )
                .about( "Generate a Sega Megadrive VDP tileset + palette from a 15-colour image" )
                .arg_from_usage( "-i, --input=<FILE> 'Specify input image'" )
                .arg_from_usage( "-o, --output=<FILE> 'Specify output file'")
                .arg(
                    Arg::with_name( "FORMAT" )
                        .short( "f" )
                        .long( "format" )
                        .help( "Specify output format for tileset (valid options: bin, inc)")
                        .default_value( "bin" )
                        .takes_value( true )
                )
                .arg(
                    Arg::with_name( "TILEORDER" )
                        .short( "to" )
                        .long( "tile-order" )
                        .help( "Specify tile order for tileset (valid options: tile, sprite" )
                        .default_value( "tile" )
                        .takes_value( true )
                )
        )
        .subcommand(
            SubCommand::with_name( "soundtrack" )
                .about( "Generate an on-console soundtrack from a chiptune tracker module" )
                .arg_from_usage( "-i, --input=<FILE> 'Specify input module'" )
                .arg_from_usage( "-o, --output=<FILE> 'Specify output filename (may output multiple files depending on format)'")
                .arg(
                    Arg::with_name( "TRACKER_FORMAT" )
                        .long( "tracker-format" )
                        .help( "Specify tracker module format (valid options: dmf (DefleMask Tracker))")
                        .default_value( "dmf" )
                        .takes_value( true )
                )
                .arg(
                    Arg::with_name( "DRIVER_FORMAT" )
                        .long( "driver-format" )
                        .help( "Specify on-console sound driver format (valid options: esf (Echo))")
                        .default_value( "esf" )
                        .takes_value( true )
                )
                .arg(
                    Arg::with_name( "EXPORT_INSTRUMENT_LIST_LABEL" )
                        .long( "instrument-list-label" )
                        .help( "In sound drivers supporting generated-code instrument lists, specify label to use for instrument list" )
                        .default_value( "MuzInstrumentList" )
                        .takes_value( true )
                )
                .arg(
                    Arg::with_name( "EXPORT_INSTRUMENT_LIST_FORMAT" )
                        .long( "instrument-list-format" )
                        .help( "In sound drivers supporting generated-code instrument lists, specify output format for instrument list" )
                        .default_value( "asmx" )
                        .takes_value( true )
                )
        )
        .get_matches();

    if let Some( matches ) = matches.subcommand_matches( "tileset" ) {
        let input_filename = matches.value_of( "input" ).ok_or( "expected: input filename (-i,--input)" )?;
        let output_filename = matches.value_of( "output" ).ok_or( "expected: output_filename (-o,--output)" )?;

        tileset::generate(
            input_filename,
            output_filename,
            matches.value_of( "FORMAT" ).unwrap(),
            matches.value_of( "TILEORDER" ).unwrap()
        );

        Ok( () )
    } else if let Some( matches ) = matches.subcommand_matches( "soundtrack" ) {
        let input_filename = matches.value_of( "input" ).ok_or( "expected: input filename (-i,--input)" )?;
        let output_filename = matches.value_of( "output" ).ok_or( "expected: output_filename (-o,--output)" )?;

        let result = DmfModule::from_file( &input_filename )?;

        for ( filename, data ) in result.get_instruments( matches.value_of( "EXPORT_INSTRUMENT_LIST_LABEL" ).ok_or( "internal error" )? )? {
            println!( "Creating instrument file {}", filename );
            let mut file = File::create( filename )?;
            file.write_all( &data )?
        }

        println!( "Creating ESF output file {}", output_filename );
        let mut file = File::create( output_filename )?;
        file.write_all( &result.get_esf()? )?;

        print_good( &format!( "successfully compiled soundtrack for {}", output_filename ) );

        Ok( () )
    } else {
        Err( "no tool provided (try --help)" )?
    }
}

fn main() {
    if let Err( error ) = run() {
        utility::print_error( &format!( "{}", error ) );
    }
}