use std::{collections::HashMap, error::Error};
use linked_hash_set::LinkedHashSet;

use crate::reskit::{soundtrack::{types::{PatternRow, Note, Channel, OctaveFrequency, Effect}, formats::dmf::DmfModule}, utility::print_warning};

// https://github.com/sikthehedgehog/Echo/blob/master/doc/esf.txt
const ESF_NOTE_ON: u8 = 0x00;
const ESF_NOTE_OFF: u8 = 0x10;
const ESF_SET_VOLUME: u8 = 0x20;
const ESF_SET_FREQUENCY: u8 = 0x30;
const ESF_SET_INSTRUMENT: u8 = 0x40;
const ESF_DELAY_SHORT: u8 = 0xD0;
const ESF_SFX_LOCK: u8 = 0xE0;
const ESF_SET_FM_PARAMS: u8 = 0xF0;
const ESF_SET_FM_DIRECT_0: u8 = 0xF8;
const ESF_SET_FM_DIRECT_1: u8 = 0xF9;
const ESF_SET_FLAGS: u8 = 0xFA;
const ESF_CLEAR_FLAGS: u8 = 0xFB;
const ESF_GOTO: u8 = 0xFC;
const ESF_LOOP_SET: u8 = 0xFD;
const ESF_DELAY_LONG: u8 = 0xFE;
const ESF_STOP: u8 = 0xFF;

pub const ESF_FM_1: u8 = 0x00;
pub const ESF_FM_2: u8 = 0x01;
pub const ESF_FM_3: u8 = 0x02;
pub const ESF_FM_4: u8 = 0x04;
pub const ESF_FM_5: u8 = 0x05;
pub const ESF_FM_6: u8 = 0x06;

pub const ESF_PSG_1: u8 = 0x08;
pub const ESF_PSG_2: u8 = 0x09;
pub const ESF_PSG_3: u8 = 0x0A;
pub const ESF_PSG_4: u8 = 0x0B;

pub const ESF_FM_6_PCM: u8 = 0x0C;

const ESF_FM_PARAM_LEFT_SPEAKER: u8 = 0x80;
const ESF_FM_PARAM_RIGHT_SPEAKER: u8 = 0x40;

const ESF_SEMITONE_C: u8        = 0;
const ESF_SEMITONE_C_SHARP: u8  = 1;
const ESF_SEMITONE_D: u8        = 2;
const ESF_SEMITONE_D_SHARP: u8  = 3;
const ESF_SEMITONE_E: u8        = 4;
const ESF_SEMITONE_F: u8        = 5;
const ESF_SEMITONE_F_SHARP: u8  = 6;
const ESF_SEMITONE_G: u8        = 7;
const ESF_SEMITONE_G_SHARP: u8  = 8;
const ESF_SEMITONE_A: u8        = 9;
const ESF_SEMITONE_A_SHARP: u8  = 10;
const ESF_SEMITONE_B: u8        = 11;

pub type EchoEvent = Vec<u8>;

pub trait EchoFormat {

    fn get_instruments( &self, instrument_list_label: &str ) -> Result<HashMap<String, Vec<u8>>, Box<dyn Error>>;

    fn get_esf( &self ) -> Result<Vec<u8>, Box<dyn Error>>;

}

fn get_semitone( note: &Note ) -> Result<u8, Box<dyn Error>> {
    Ok( match note {
        Note::NoteOff => return Err( "internal error: attempted to get semitone for a Note::NoteOff" )?,
        Note::C(_) => ESF_SEMITONE_C,
        Note::CSharp(_) => ESF_SEMITONE_C_SHARP,
        Note::D(_) => ESF_SEMITONE_D,
        Note::DSharp(_) => ESF_SEMITONE_D_SHARP,
        Note::E(_) => ESF_SEMITONE_E,
        Note::F(_) => ESF_SEMITONE_F,
        Note::FSharp(_) => ESF_SEMITONE_F_SHARP,
        Note::G(_) => ESF_SEMITONE_G,
        Note::GSharp(_) => ESF_SEMITONE_G_SHARP,
        Note::A(_) => ESF_SEMITONE_A,
        Note::ASharp(_) => ESF_SEMITONE_A_SHARP,
        Note::B(_) => ESF_SEMITONE_B
    } )
}

/**
 * Notes are used to index PCM sample playback. C = 1, etc.
 */
fn get_pcm_index( note: &Note ) -> Result<u8, Box<dyn Error>> {
    Ok( match note {
        Note::NoteOff => return Err( "internal error: attempted to get pcm index for a Note::NoteOff" )?,
        Note::C(_) => 0,
        Note::CSharp(_) => 1,
        Note::D(_) => 2,
        Note::DSharp(_) => 3,
        Note::E(_) => 4,
        Note::F(_) => 5,
        Note::FSharp(_) => 6,
        Note::G(_) => 7,
        Note::GSharp(_) => 8,
        Note::A(_) => 9,
        Note::ASharp(_) => 10,
        Note::B(_) => 11
    } )
}

/**
  Gets the semitone frequency used by Echo.
  https://github.com/sikthehedgehog/Echo/blob/master/doc/esf.txt#L243
    C  - 644 | E  - 810 | G# - 1021
    C# - 681 | F  - 858 | A  - 1081
    D  - 722 | F# - 910 | A# - 1146
    D# - 765 | G  - 964 | B  - 1214
*/
fn get_semitone_frequency( semitone: u8 ) -> Result<u16, Box<dyn Error>> {
    Ok( match semitone {
        ESF_SEMITONE_C => 644,
        ESF_SEMITONE_C_SHARP => 681,
        ESF_SEMITONE_D => 722,
        ESF_SEMITONE_D_SHARP => 765,
        ESF_SEMITONE_E => 810,
        ESF_SEMITONE_F => 858,
        ESF_SEMITONE_F_SHARP => 910,
        ESF_SEMITONE_G => 964,
        ESF_SEMITONE_G_SHARP => 1021,
        ESF_SEMITONE_A => 1081,
        ESF_SEMITONE_A_SHARP => 1146,
        ESF_SEMITONE_B => 1214,
        _ => return Err( "internal error: invalid semitone value provided to get_semitone_frequency" )?
    } )
}

fn get_fm_note_byte( octave: u8, semitone: u8 ) -> u8 {
    32 * octave + 2 * semitone + 1
}

fn get_psg_note_byte( octave: u8, semitone: u8 ) -> u8 {
    24 * octave + 2 * semitone
}

fn get_note( note: &Note, channel: u8 ) -> Result<u8, Box<dyn Error>> {
    Ok( match channel {
        ESF_FM_1..=ESF_FM_3 | ESF_FM_4..=ESF_FM_6 => get_fm_note_byte( note.get_octave()?, get_semitone( note )? ),
        ESF_PSG_1..=ESF_PSG_4 => get_psg_note_byte( note.get_octave()?, get_semitone( note )? ),
        invalid_channel => return Err( format!( "internal error: get_note: invalid channel {:#04X}", invalid_channel ) )?
    } )
}

fn get_volume( channel: u8, volume: u8 ) -> Result<u8, Box<dyn Error>> {
    Ok( match channel {
        ESF_FM_1..=ESF_FM_3 | ESF_FM_4..=ESF_FM_6 => 0x7F - volume,
        ESF_PSG_1..=ESF_PSG_4 => 0x0F - volume,
        invalid_channel => return Err( format!( "internal error: get_note: invalid channel {:#04X}", invalid_channel ) )?
    } )
}

/**
 * Generate an echo delay event. "delay" is the amount to wait, in 1/60 of a second ticks.
 */
fn get_delay( delay: u8 ) -> Result<EchoEvent, Box<dyn Error>> {
    let events: EchoEvent = if delay == 0 {
        // No delay to generate! Probably should be an error
        vec![]
    } else if delay <= 16 {
        vec![ ESF_DELAY_SHORT | ( delay - 1 ) ]
    } else {
        vec![ ESF_DELAY_LONG, delay ]
    };

    Ok( events )
}

/**
 * Generate an Echo frequency shift up event for a given channel, updating active_note to be ready
 * for the next event.
 */
fn get_portamento( channel: &mut Channel, portamento_effect: &Effect ) -> Result<EchoEvent, Box<dyn Error>> {
    Ok(
        if channel.id == ESF_FM_6_PCM {
            // Generate no events if channel is set to pcm
            print_warning( "attempted to portamento up on FM6 when it is set to pcm. this makes no sense and does nothing" );
            vec![]
        } else if let Some( octave_frequency ) = channel.active_note {
            let ( octave, frequency ) = {
                let by_amount: i16 = match portamento_effect {
                    Effect::PortamentoUp { speed } => *speed as i16,
                    Effect::PortamentoDown { speed } => -( *speed as i16 ),
                    _ => return Err( "internal error: provided effect is not a supported portamento effect" )?
                };

                let new_frequency: i16 = octave_frequency.frequency as i16 + by_amount;
                if new_frequency > get_semitone_frequency( ESF_SEMITONE_B )? as i16 {
                    if octave_frequency.octave == 7 {
                        // Nowhere else to go up
                        ( 7, get_semitone_frequency( ESF_SEMITONE_B )? )
                    } else {
                        // Go up an octave then add the difference to middle C
                        let difference = new_frequency - get_semitone_frequency( ESF_SEMITONE_B )? as i16;
                        ( octave_frequency.octave + 1, get_semitone_frequency( ESF_SEMITONE_C )? + difference as u16 )
                    }
                } else if new_frequency < get_semitone_frequency( ESF_SEMITONE_C )? as i16 {
                    if octave_frequency.octave == 0 {
                        // Nowhere else to go down
                        ( 0, get_semitone_frequency( ESF_SEMITONE_C )? )
                    } else {
                        // Go down an octave then subtract the overshoot of C from B
                        let difference = get_semitone_frequency( ESF_SEMITONE_C )? as i16 - new_frequency;
                        ( octave_frequency.octave - 1, get_semitone_frequency( ESF_SEMITONE_B )? - difference as u16 )
                    }
                } else {
                    // Move within the same octave
                    ( octave_frequency.octave, new_frequency as u16 )
                }
            };

            // Set the new OctaveFrequency on the channel
            channel.active_note = Some( OctaveFrequency { octave, frequency } );

            // Generate the note slide ESF events
            // YM2612 register format for frequency: https://plutiedev.com/ym2612-registers#reg-A0
            vec![
                ESF_SET_FREQUENCY | channel.id,
                octave << 3 | ( ( ( 0x0700 & frequency ) >> 8 ) as u8 ),
                ( 0x00FF & frequency ) as u8
            ]
        } else {
            // No active note, nothing to generate
            vec![]
        }
    )
}

/**
 * For a specific row and channel, get the events this row and channel contribute to the stream.
 * While doing so, update the state of the channel for future event generation.
 */
fn get_events_for_channel( channel: &mut Channel, row: &PatternRow, pcm_offset: u8 ) -> Result<Vec<EchoEvent>, Box<dyn Error>> {
    let mut events: Vec<EchoEvent> = Vec::new();

    // Adjust volume
    if let Some( volume ) = &row.volume {
        // https://github.com/sikthehedgehog/Echo/blob/master/doc/esf.txt#L206-L207
        let volume = get_volume( channel.id, *volume )?;
        channel.current_volume = Some( volume );
        events.push( vec![ ESF_SET_VOLUME | channel.id, volume ] );
    }

    // Set the instrument
    if let Some( instrument ) = &row.instrument_index {
        // Do not set the same instrument if it was already set before
        if &row.instrument_index != &channel.active_instrument {
            events.push( vec![ ESF_SET_INSTRUMENT | channel.id, *instrument as u8 ] );
            channel.active_instrument = Some( *instrument );
        }
    }

    // Key on or key off note
    if let Some( note ) = &row.note {
        if let Note::NoteOff = note {
            channel.active_note = None;
            events.push( vec![ ESF_NOTE_OFF | channel.id ] );
        } else {
            channel.active_note = Some(
                OctaveFrequency {
                    octave: note.get_octave()?,
                    frequency: get_semitone_frequency( get_semitone( note )? )?
                }
            );

            if channel.id == ESF_FM_6_PCM {
                events.push( vec![ ESF_NOTE_ON | channel.id, get_pcm_index( note )? + pcm_offset ] );
            } else {
                let note = get_note( note, channel.id )?;
                events.push( vec![ ESF_NOTE_ON | channel.id, note ] );
            }
        }
    }

    Ok( events )
}

/**
 * Add effects from a given row to a particular channel and perform activities due on insert. This may generate Echo
 * events that should be applied to the stream.
 */
fn apply_effects_to_channel( channel: &mut Channel, effects: &LinkedHashSet<Effect> ) -> Result<Vec<EchoEvent>, Box<dyn Error>> {
    let mut events: Vec<EchoEvent> = Vec::new();

    for effect in effects {
        match effect {
            Effect::DacEnable { enabled } => {
                if !enabled {
                    if channel.id == ESF_FM_6 || channel.id == ESF_FM_6_PCM {
                        channel.id = ESF_FM_6;
                    } else {
                        print_warning( "Effect::DacEnable can only be applied to FM6. ignoring..." );
                    }

                    channel.active_effects.remove( &Effect::DacEnable { enabled: true } );
                } else {
                    if channel.id == ESF_FM_6 || channel.id == ESF_FM_6_PCM {
                        channel.id = ESF_FM_6_PCM;
                    } else {
                        print_warning( "Effect::DacEnable can only be applied to FM6. ignoring..." );
                    }

                    channel.active_effects.insert_if_absent( Effect::DacEnable { enabled: true } );
                }
            },
            Effect::PortamentoUp { speed } => {
                let speed = *speed;

                // Remove all prior portamentos (you cannot portamento up and down at the same time)
                channel.active_effects = channel.active_effects
                                                .clone()
                                                .into_iter()
                                                .filter( | effect |
                                                    !(
                                                        matches!( effect, Effect::PortamentoDown { speed: _ } ) ||
                                                        matches!( effect, Effect::PortamentoUp { speed: _ } )
                                                    )
                                                )
                                                .collect();

                if speed > 0 {
                    channel.active_effects.insert( Effect::PortamentoUp { speed } );
                }
            },
            Effect::PortamentoDown { speed } => {
                let speed = *speed;

                // Remove all prior portamentos (you cannot portamento up and down at the same time)
                channel.active_effects = channel.active_effects
                                                .clone()
                                                .into_iter()
                                                .filter( | effect |
                                                    !(
                                                        matches!( effect, Effect::PortamentoDown { speed: _ } ) ||
                                                        matches!( effect, Effect::PortamentoUp { speed: _ } )
                                                    )
                                                )
                                                .collect();

                if speed > 0 {
                    channel.active_effects.insert( Effect::PortamentoDown { speed } );
                }
            },
            Effect::NoteCut { after_ticks } => {
                let after_ticks = *after_ticks;

                // Remove all previous note cuts
                channel.active_effects = channel.active_effects
                                                .clone()
                                                .into_iter()
                                                .filter( | effect | !( matches!( effect, Effect::NoteCut { after_ticks: _ } ) ) )
                                                .collect();

                // Add new note cut. Being an effect requiring delay generation, it will generate (and expire) in get_delays
                channel.active_effects.insert( Effect::NoteCut { after_ticks } );
            }
            Effect::SetPanning { left, right } => {
                let left = *left;
                let right = *right;

                if channel.id == ESF_FM_6_PCM || ( ESF_PSG_1..=ESF_PSG_4 ).contains( &channel.id ) {
                    print_warning( "Effect::SetPanning unsupported on PCM or PSG channel" );
                } else {
                    if !channel.active_effects.contains( &Effect::SetPanning { left, right } ) {
                        // Remove other active SetPanning effects
                        channel.active_effects.remove( &Effect::SetPanning { left: true, right: true } );
                        channel.active_effects.remove( &Effect::SetPanning { left: true, right: false } );
                        channel.active_effects.remove( &Effect::SetPanning { left: false, right: true } );
                        channel.active_effects.remove( &Effect::SetPanning { left: false, right: false } );

                        // Insert new panning commands
                        let panning: u8 = if left && right {
                            ESF_FM_PARAM_LEFT_SPEAKER & ESF_FM_PARAM_RIGHT_SPEAKER
                        } else if left {
                            ESF_FM_PARAM_LEFT_SPEAKER
                        } else if right {
                            ESF_FM_PARAM_RIGHT_SPEAKER
                        } else {
                            0x00
                        };
                        events.push( vec![ ESF_SET_FM_PARAMS | channel.id, panning ] );

                        // Insert new active SetPanning effect - will not be doubly added
                        channel.active_effects.insert( Effect::SetPanning { left, right } );
                    }
                }
            },
            unsupported_effect => print_warning( &format!( "effect unsupported: {:?}. your soundtrack may sound different than expected.", unsupported_effect ) )
        }
    }

    Ok( events )
}

/**
 * Replace concurrent sequences of waits with single larger waits. This reduces the ESF stream size.
 */
pub fn compact_delays( events: Vec<EchoEvent> ) -> Result<Vec<EchoEvent>, Box<dyn Error>> {
    let mut new_events: Vec<EchoEvent> = Vec::new();

    let mut current_sequence: Vec<EchoEvent> = Vec::new();
    for event in events {
        if let Some( esf_event_type ) = event.iter().next() {
            if *esf_event_type & 0xF0 == ESF_DELAY_SHORT || *esf_event_type == ESF_DELAY_LONG {
                // Continue to push delay events into current_sequence
                current_sequence.push( event );
            } else {
                // Encountered a non-delay event - we will always push something here,
                // but let's check first if we need to push a current_sequence of waits.
                // Coalesce all current delays in current_sequence and push them
                // onto the new event stream.
                let mut cumulative_delay: u16 = 0;
                for delay_event in &current_sequence {
                    let mut iter = delay_event.iter();
                    let delay_type = *iter.next().ok_or( "internal error: empty event" )?;

                    if delay_type & 0xF0 == ESF_DELAY_SHORT {
                        cumulative_delay += ( ( delay_type & 0x0F ) as u16 ) + 1;
                    } else if delay_type == ESF_DELAY_LONG {
                        cumulative_delay += ( *iter.next().ok_or( "internal error: expected delay byte" )? ) as u16;
                    }
                }

                if cumulative_delay > 0 {
                    // Push amount of full-256 wait events
                    for _ in 0..cumulative_delay / 256 {
                        new_events.push( vec![ ESF_DELAY_LONG, 0x00 ] );
                    }

                    // Push the remainder under 256 event
                    new_events.push( get_delay( ( cumulative_delay % 256 ) as u8 )? );

                    // Reset the current sequence
                    current_sequence.clear();
                }

                new_events.push( event );
            }
        }
    }

    Ok( new_events )
}

/**
 * Calculate delays generated before and after effects applied across all channels. If there are no effects for this row,
 * an empty vec will be returned. This means you should just apply ticks_to_wait as the delay for that row. If there
 * -are- effects generated by this function, you should not apply ticks_to_wait, since this function will spend out
 * the tick budget for that row instead.
 */
fn get_delays_for_effects( channels: &mut [Channel], ticks_to_wait: u8 ) -> Result<Vec<EchoEvent>, Box<dyn Error>> {
    let mut events: Vec<EchoEvent> = Vec::new();

    // All portamento effects deploy per tick, not per row. So we need to aggregate all portamentos across all
    // channels for this row, then flush them once per `ticks_to_wait` for this row.
    let mut active_portamentos: Vec<(usize, Effect)> = Vec::new();
    for channel_id in 0..channels.len() {
        let channel = &channels[ channel_id ];

        for effect in &channel.active_effects {
            if matches!( effect, Effect::PortamentoDown { speed: _ } ) ||
               matches!( effect, Effect::PortamentoUp { speed: _ } ) {
                active_portamentos.push( ( channel_id, *effect ) );
            }
        }
    }

    if !active_portamentos.is_empty() {
        for _tick in 0..ticks_to_wait {
            for ( channel_id, portamento ) in &active_portamentos {
                let portamento_event: EchoEvent = get_portamento( &mut channels[ *channel_id ], &portamento )?;
                events.push( portamento_event );
            }

            events.push( get_delay( 1 )? );
        }
    }

    Ok( events )
}

/**
 * Get the delays due at the end of a row. These delays are what flushes the tick to Echo so that it can play.
 */
fn get_delays( events: Vec<EchoEvent>, channels: &mut [Channel], ticks_to_wait: u8 ) -> Result<Vec<EchoEvent>, Box<dyn Error>> {
    let mut events: Vec<EchoEvent> = events;

    let applied_effects = get_delays_for_effects( channels, ticks_to_wait )?;
    if applied_effects.is_empty() {
        // Push the amount of ticks to wait for this row. ticks_to_wait is speed_a or speed_b, times base_speed.
        events.push( get_delay( ticks_to_wait )? );
    } else {
        // Do not push the default "wait n ticks" event but rather push the delay events generated by effects
        events.extend( applied_effects.into_iter() );
    }

    Ok( events )
}

/**
 * For an entire row across all channels, generate the events that apply to the row as a whole. This usually means applying
 * the waits so that the row can be flushed to Echo and played - ESF ticks are until the nearest wait or stop event.
 */
pub fn get_events_for_row( channels: &mut [Channel], subrows: Vec<&PatternRow>, ticks_to_wait: u8, pcm_offset: u8 ) -> Result<Vec<EchoEvent>, Box<dyn Error>> {
    let mut events: Vec<EchoEvent> = Vec::new();

    // Get events for each subrow (part of the total row for each channel)
    for i in 0..channels.len() {
        // Apply effects to channel's current state
        events.extend( apply_effects_to_channel( &mut channels[ i ], &subrows[ i ].effects )? );

        // Get the ESF events for this channel's part of the row
        events.extend( get_events_for_channel( &mut channels[ i ], subrows[ i ], pcm_offset )? );
    }

    Ok( get_delays( events, channels, ticks_to_wait )? )
}